#include "wavefront.h"
#include <QPainter>
#include <QApplication>
#include <QPushButton>
#include <QTextStream>
#include <cmath>
#include <QLineEdit>
#include <QLabel>
#include <QString>
#include <iostream>
#include <QTextStream>
#include <QMenu>
#include <QMenuBar>
#include <QMainWindow>
#include <QTime>
#include <QSlider>
#include <QRadioButton>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QCursor>
#include <QCheckBox>

#ifndef spat_dist
#define spat_dist 512
#endif

#define OFFSETH 10

using namespace std;

wavefront::wavefront(QWidget * parent)
    : QMainWindow(parent)
{
  iter=0;
  virtual_time=0;
  //this->setStyleSheet("background-color: gray;");


  //MENU BAR
  QAction *quit = new QAction("&Quit", this);

  QMenu *file;
  
  file = menuBar()->addMenu("&File");
  file->addAction(quit);

  static_noise_is_present=false;
  noise_intencity=0;

  connect(quit, SIGNAL(triggered()), qApp, SLOT(quit()));
  

  /* VISUALS ON/OFF ARCHITECTURE */
  visuals=1;
  visualsBox = new QCheckBox("Visualization on/off ", this);
  visualsBox->setGeometry(170, OFFSETH+660, 230, 30);
  visualsBox->setChecked(1);
  connect(visualsBox, SIGNAL(clicked()), this, SLOT(setVisuals()));

  //STATIC NOISE ARCHITECURE
  noise_intencity_slider = new QSlider(Qt::Horizontal , this);
  noise_intencity_slider->setGeometry(590, OFFSETH+560, 230, 30);
  noise_intencity_label = new QLabel("sg: ", this);
  noise_intencity_label->setGeometry(560, OFFSETH+560, 30, 30);
  noise_intencity_label2 = new QLabel("0", this);
  noise_intencity_label2->setGeometry(825, OFFSETH+560, 30, 30);
  static_noise_button = new QPushButton("add static noise", this);
  static_noise_button->setGeometry(560, OFFSETH+590, 300, 30);
  noise_intencity_slider->setRange(0, 600);
  connect(noise_intencity_slider, SIGNAL(valueChanged(int)), noise_intencity_label2, SLOT(setNum(int)));
  connect(noise_intencity_slider, SIGNAL(valueChanged(int)), this, SLOT(noise_intencity_alterer()));
  connect(static_noise_button, SIGNAL(clicked()), this, SLOT(set_static_noise()));

  /*  DYNAMIC NOISE ARCHITECTURE  */
  M_slider = new QSlider(Qt::Horizontal , this);
  M_slider->setGeometry(590, OFFSETH+640, 230, 30);
  M_slider->setRange(0, spat_dist);
  M_slider->setValue(0);
  M_label = new QLabel("M: ", this);
  M_label->setGeometry(560, OFFSETH+640, 30, 30);
  M_label2 = new QLabel("0", this);
  M_label2->setGeometry(825, OFFSETH+640, 30, 30);
  connect(M_slider, SIGNAL(valueChanged(int)), M_label2, SLOT(setNum(int)));
  connect(M_slider, SIGNAL(valueChanged(int)), this, SLOT(M_alterer()));

  // TIME STEP ALTERNATION ARCHITECURE
  time_step_slider = new QSlider(Qt::Horizontal , this);
  time_step_slider->setGeometry(590, OFFSETH+620, 230, 30);
  time_step_slider->setRange(1, 200);
  time_step_slider->setValue(80);
  time_label = new QLabel("dt: ", this);
  time_label->setGeometry(560, OFFSETH+620, 30, 30);
  time_label2 = new QLabel("80", this);
  time_label2->setGeometry(825, OFFSETH+620, 30, 30);
  connect(time_step_slider, SIGNAL(valueChanged(int)), time_label2, SLOT(setNum(int)));
  connect(time_step_slider, SIGNAL(valueChanged(int)), this, SLOT(time_step_alterer()));

  //START SIMULATION BUTTON
  start_button = new QPushButton("Start", this);
  start_button->setGeometry(450, OFFSETH+560, 100, 30);
  connect(start_button, SIGNAL(clicked()), this, SLOT(startWave()));

  //CHANGE VIEW BUTTON
  view_button = new QPushButton("change to v", this);
  view_button->setGeometry(450, OFFSETH+640, 100, 30);
  connect(view_button, SIGNAL(clicked()), this, SLOT(changeView()));
  
  running = 0;
  space = new wavespace();
  disp = new phasedisp();

  view=0; //0=u,1=v, 2=u/v
  space->image_x=20;
  space->image_y=OFFSETH+20;

  //CHANGE PARAMETER ARCHITECURE
  QLabel *eps_label = new QLabel("eps:", this);
  eps_label->setGeometry(50, OFFSETH+560, 60, 30);
  eps_edit = new QLineEdit("0.03",this);
  eps_edit->text();
  eps_edit->setGeometry(100, OFFSETH+560, 60, 30);

  QLabel *b_label = new QLabel("b:", this);
  b_label->setGeometry(50, OFFSETH+590, 60, 30);
  b_edit = new QLineEdit("0.001",this);
  b_edit->text();
  b_edit->setGeometry(100, OFFSETH+590, 60, 30);

  QLabel *a_label = new QLabel("a:", this);
  a_label->setGeometry(50, OFFSETH+620, 60, 30);
  a_edit = new QLineEdit("0.75",this);
  a_edit->text();
  a_edit->setGeometry(100, OFFSETH+620, 60, 30);

  QLabel *D_label = new QLabel("D:", this);
  D_label->setGeometry(50, OFFSETH+650, 60, 30);
  D_edit = new QLineEdit("1",this);
  D_edit->text();
  D_edit->setGeometry(100, OFFSETH+650, 60, 30);

  QLabel *Dv_label = new QLabel("Dv:", this);
  Dv_label->setGeometry(50, OFFSETH+680, 60, 30);
  Dv_edit = new QLineEdit("0",this);
  Dv_edit->text();
  Dv_edit->setGeometry(100, OFFSETH+680, 60, 30);

  QLabel *L_label = new QLabel("L:", this);
  L_label->setGeometry(50, OFFSETH+710, 60, 30);
  L_edit = new QLineEdit("60",this);
  L_edit->text();
  L_edit->setGeometry(100, OFFSETH+710, 60, 30);

  new_para = new QPushButton("Change parameters", this);
  new_para->setGeometry(170, OFFSETH+560, 160, 30);
  connect(new_para, SIGNAL(clicked()), this, SLOT(setparam()));

  /*  ANALYZIS VISUALIZATION ARCHITECTURE  */		
  average_iter_rate_label = new QLabel("a.it.r:  1/s", this);
  average_iter_rate_label->setGeometry(170, OFFSETH+590, 250, 35);
  elapsed_real_time = new QLabel("Elapsed real time: 0Ms", this);
  elapsed_real_time->setGeometry(170, OFFSETH+615, 250, 35);
  elapsed_virtual_time = new QLabel("Elapsed virtual time: 0", this);
  elapsed_virtual_time->setGeometry(170, OFFSETH+635, 250, 35);

  /* METHOD ALTERNATION ARCHITECTURE */
	//solution_method="implicit euler";
	solution_method="implicit euler";
	change_methods = new QPushButton("Implicit Euler", this);
	change_methods->setGeometry(450, OFFSETH+685, 220, 30);
	QMenu *methods_menu = new QMenu();
	change_methods->setMenu(methods_menu);
	QAction *exeuler = new QAction("Explicit Euler method", this);
	methods_menu->addAction(exeuler);
    connect(exeuler, SIGNAL(triggered()), this, SLOT(set_methodee()));
	QAction *impleuler = new QAction("Implicit Euler method", this);
	methods_menu->addAction(impleuler);
    connect(impleuler, SIGNAL(triggered()), this, SLOT(set_methodie()));
	QAction *shardlow = new QAction("Shardlow's scheme", this);
	methods_menu->addAction(shardlow);
    connect(shardlow, SIGNAL(triggered()), this, SLOT(set_methodsh()));
	QAction *eulerlawson = new QAction("Lawson-Euler", this);
	methods_menu->addAction(eulerlawson);
    connect(eulerlawson, SIGNAL(triggered()), this, SLOT(set_methodel()));
	QAction *IFRK4 = new QAction("IFRK4 (custom problem)", this);
	methods_menu->addAction(IFRK4);
    connect(IFRK4, SIGNAL(triggered()), this, SLOT(set_methodIFRK4()));

  /* REACTION KINETCS ALTERNATION ARCHITECTURE */
	reaction_kinetics="standard";
	change_reaction_kinetics = new QPushButton("Standard reaction kinetics", this);
	change_reaction_kinetics->setGeometry(450, OFFSETH+725, 220, 30);
	QMenu *reaction_menu = new QMenu();
	change_reaction_kinetics->setMenu(reaction_menu);
	QAction *stdrea = new QAction("Standard reaction kinetics", this);
	reaction_menu->addAction(stdrea);
    connect(stdrea, SIGNAL(triggered()), this, SLOT(set_reactionstd()));
	QAction *delay = new QAction("Delayed reaction kinetics", this);
	reaction_menu->addAction(delay);
    connect(delay, SIGNAL(triggered()), this, SLOT(set_reactiondel()));
	QAction *bareisw = new QAction("bareisw reaction kinetics", this);
	reaction_menu->addAction(bareisw);
    connect(bareisw, SIGNAL(triggered()), this, SLOT(set_reactionbareisw()));

  /* RESET WAVE FIELD ARCHITECTURE */
	reset_button = new QPushButton("reset", this);
	reset_button->setGeometry(450, OFFSETH+600, 100, 30);
	QMenu * reset_menu = new QMenu();
	reset_button->setMenu(reset_menu);
	QAction * spiral = new QAction("spiral wave", this);
	reset_menu->addAction(spiral);
	connect(spiral, SIGNAL(triggered()), this, SLOT(resetWave()));
	QAction * pointstim = new QAction("point stimuli", this);
	reset_menu->addAction(pointstim);
	connect(pointstim, SIGNAL(triggered()), this, SLOT(resetPointstim()));
	QAction * clear = new QAction("clear wave", this);
	reset_menu->addAction(clear);
	connect(clear, SIGNAL(triggered()), this, SLOT(resetClear()));
}

wavefront::~wavefront()
{
  //DESTRUCTOR
}

void wavefront::paintEvent(QPaintEvent *event)
{

	if (visuals)
	{
	  QPainter qpainter(this);

	  QRgb value;

	  for (int xx=0; xx<(spat_dist); xx++)
	  {
		for (int yy=0; yy<(spat_dist); yy++)
		{
			if (!space->inexfield[xx][yy])
			{
				  if (view==0) value = qRgb(0, 0, 250*abs(space->getu(xx,yy)));
				  else if (view==2) value = qRgb(0, 250*abs(space->getv(xx,yy)), 250*abs(space->getu(xx,yy)));
				  else value = qRgb(0, 250*abs(space->getv(xx,yy)),0);
			}
			else {value = qRgb(250, 250, 250);};
		      space->image->setPixel(xx, yy, value);
		}
	  }

	  qpainter.drawImage(space->image_x, space->image_y, *space->image);
	  qpainter.drawImage(disp->imgx, OFFSETH+ disp->imgy, *disp->image);
	  qpainter.drawImage(560, OFFSETH+324, *disp->plot);
	}
}

void wavefront::timerEvent(QTimerEvent *event)
{

  QTime qtime;
  qtime.start();
  if (running)
  {
	iter++;
	set_temporal_info();
	
    //space->add_whoise(noise_intencity,iter);

	if (solution_method=="explicit euler" || solution_method=="implicit euler")
	{
    	space->update_space(noise_intencity,iter,solution_method,reaction_kinetics);
	}
	else if (solution_method=="shardlow's scheme")
	{
    	space->spectral_update_space(noise_intencity,iter,reaction_kinetics);
	}
	else if (solution_method=="euler-lawson")
	{
		space->eulerlawson_update_space(noise_intencity,iter,reaction_kinetics);
	}
	else if (solution_method=="IFRK4")
	{
		space->IFRK4(noise_intencity,iter);
	}
    
	disp->update_vec(space->getu(disp->x,disp->y),space->getv(disp->x,disp->y));    
    disp->update_image();
    disp->update_plot(); 
  }
/*
	if (iter==200)
	{
		running=0;
		//space->M=512;
		//noise_intencity=90;
	}

	if(iter==230)
  	{
		space->image->save("sg0T230.png",0,-1);
  	}

	if(iter==290)
  	{
		space->image->save("sg0T290.png",0,-1);
  	}

	if(iter==470)
  	{
		space->image->save("sg0T470.png",0,-1);
  	}

	if(iter==540)
  	{
		space->image->save("sg0T540.png",0,-1);
  	}

	if(iter==720)
  	{
		space->image->save("sg0T720.png",0,-1);
  	}

	if(iter==1120)
  	{
		space->image->save("noiseat20T112.png",0,-1);
  	}
*/
/*
	if(iter==620)
  	{
		space->image->save("labyrinthine62.jpg",0,-1);
  	}

	if(iter==1000)
  	{
		space->image->save("labyrinthine100.jpg",0,-1);
  	}


  if(iter==200)
  {
		space->image->save("image200.jpg",0,-1);
  }

  if(iter==600)
  {
		space->image->save("image600.jpg",0,-1);
  }

  if(iter==1000)
  {
		space->image->save("image1000.jpg",0,-1);
  }
*/
  repaint();
}

void wavefront::mousePressEvent(QMouseEvent *event)
{
  if (event)
  {
    mousepos=event->pos();
    printf("x: %i y: %i \n",mousepos.x(),mousepos.y());

    // ser om du trykker på hovedvinduet:
    if (space->image_x<=mousepos.x() && mousepos.x()<=(space->image_x+spat_dist))
    {
       if (space->image_y<=mousepos.y() && mousepos.y()<=(space->image_y+spat_dist))
       {
        printf("IMAGE CLICK DETECTED ! ! !\n");
        disp->change_point(mousepos.x()-space->image_x,mousepos.y()-space->image_y);
		printf("point value: u=%f v=%f \n",space->u_space[mousepos.x()-space->image_x][mousepos.y()-space->image_y],
		space->v_space[mousepos.x()-space->image_x][mousepos.y()-space->image_y]);
		return;
       }
    }

	// ser om du trykker på faseplottet:
    if (disp->imgx<=mousepos.x() && mousepos.x()<=(disp->imgx+disp->imgw))
    {
       if (disp->imgy<=mousepos.y() && mousepos.y()<=(disp->imgy+disp->imgh))
       {
        printf("changing value\n");

		double a = (double)(mousepos.x()-disp->imgx)/disp->imgw;
		double b = 1-(double)(mousepos.y()-disp->imgy)/disp->imgh;
	
		cout << a << endl;
		cout << b << endl;
		
        space->set_value(a,b, disp->x, disp->y);

		return;
       }
    }
  }
}

void wavefront::keyPressEvent(QKeyEvent *event)
{
    switch (event->key()) {
    case Qt::Key_Left:
       {  
        
       }
	break;
    case Qt::Key_Right:
        { 
         
        }
	break;      
    case Qt::Key_S:
        { 
         space->image->save("image.jpg",0,-1);
	 	 printf("current number of iterations: %i \n",iter);
        }
	break;
    case Qt::Key_Space:
        {

        }
        break;
    case Qt::Key_Escape:
        {
          
        }
        break;
    default:
        QWidget::keyPressEvent(event);
    }
}

void wavefront::startWave()
{
  if (!running)
  {
    start_button->setText("pause");
    running = 1;
    timerId = startTimer(1);
	runTimer.startTimer();
  }
  else
  {
    start_button->setText("restart");
    running = 0;
	runTimer.pauseTimer();
  }
}

void wavefront::setparam()
{
  //set inn funksjon som sjekker om nye param er valid!
  float nyeps=eps_edit->text().toDouble();
  float nyb=b_edit->text().toDouble();
  float nya=a_edit->text().toDouble();
  float nyD=D_edit->text().toDouble();
  float nyDv=Dv_edit->text().toDouble();
  int nyL=L_edit->text().toInt();
  /*
  if (nyeps>1 || nyeps<=0) nyeps=0.03; 
  if (nyb>10 || nyb<=-10) nyb=0.006;
  if (nya>1 || nya<=0) nyeps=0.75; 
  if (nyD>20 || nyD<=0) nyD=1;
  if (nyDv>20 || nyD<=0) nyDv=1;
  if (nyL>spat_dist || nyL<=0) nyL=1;*/

  space->set_param(nyeps,nyb,nya,nyD,nyDv,nyL);
}

void wavefront::resetWave()
{
	virtual_time=0;
	runTimer.setElapsed(0);
	iter=0;

	space->inicond();
}

void wavefront::resetClear()
{
	virtual_time=0;
	runTimer.setElapsed(0);
	iter=0;

	space->clear();
}

void wavefront::resetPointstim()
{
	virtual_time=0;
	runTimer.setElapsed(0);
	iter=0;

	space->clear();
	space->u_space[spat_dist/2][spat_dist/2]=0.9;
}

void wavefront::changeView()
{
  if (view==0) // viser u
  {
    view_button->setText("change to u/v");
    view=1;
  }
  else if (view==1) // viser v
  {
    view_button->setText("change to u");
    view = 2;
  }
  else // viser både u og v
  {
    view=0;
    view_button->setText("change to v");
  }
}

void wavefront::set_static_noise()
{
	static_noise_is_present=!static_noise_is_present;

	if (static_noise_is_present)
	{
		static_noise_button->setText("remove static noise");
		space->distribute_inxtblty(1, noise_intencity);
	}
	else
	{
		static_noise_button->setText("add static noise");
		space->distribute_inxtblty(0,0);
	}
}

void wavefront::noise_intencity_alterer()
{
	noise_intencity=(double)(noise_intencity_slider->value())/10;
}

void wavefront::time_step_alterer()
{
	space->set_dt((float)(time_step_slider->value())/1000);
}

void wavefront::M_alterer()
{
	space->set_M((float)(M_slider->value()));

	//double LHS, eigM, RHS;
	//eigM = pow(3.141592/space->L,2);
	//LHS = noise_intencity*1/(8*eigM)*pow(space->L/space->spatialCorr,2)*exp(-3.141592*pow(space->M-1,2)*pow(space->spatialCorr/space->L,2));
	//RHS = pow((double)(1./spat_dist),2);
	//printf("%e < %e ?\n",LHS,RHS);
}

void wavefront::set_temporal_info()
{	
	float rate=(float)iter/(float)runTimer.getElapsedTime();

	QString qs = "a.it.r: " + QString::number(rate)+" 1/Ms";

	average_iter_rate_label->setText(qs);
	qs = "Elapsed real time: " + QString::number(runTimer.getElapsedTime())+"Ms";
	elapsed_real_time->setText(qs);

	virtual_time+=space->dt;
	qs = "Elapsed virtual time: " + QString::number(virtual_time);
	elapsed_virtual_time->setText(qs);

}

void wavefront::set_methodee()
{
	change_methods->setText("Explicit Euler");
	solution_method="explicit euler";
}

void wavefront::set_methodie()
{
	change_methods->setText("Implicit Euler");
	solution_method="implicit euler";
}

void wavefront::set_methodsh()
{
	change_methods->setText("Sharldow's scheme");
	solution_method="shardlow's scheme";
}

void wavefront::set_methodel()
{
	change_methods->setText("Lawson-Euler");
	solution_method="euler-lawson";
}

void wavefront::set_methodIFRK4()
{
	change_methods->setText("IFRK4 (custom problem)");
	solution_method="IFRK4";
}

void wavefront::set_reactionstd()
{
	change_reaction_kinetics->setText("Standard reaction kinetics");
	reaction_kinetics="standard";
}

void wavefront::set_reactiondel()
{
	change_reaction_kinetics->setText("Delayed reaction kinetics");
	reaction_kinetics="delayed";
}

void wavefront::set_reactionbareisw()
{
	change_reaction_kinetics->setText("bareisw reaction kinetics");
	reaction_kinetics="bareisw";
}

void wavefront::setVisuals()
{
	visuals=visualsBox->isChecked();
}
