#include "wavespace.h"
#include <QTextStream>
#include <QPointF>
#include <QPainter>
#include <iostream>
#include <cstdio>
#include <fftw3.h>
#include <cmath>
#include <string>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#ifndef spat_dist
#define spat_dist 512
#endif

#ifndef pi
#define pi 3.141592653589793
#endif

//SOLUTION SCHEME PARAMETERS

using namespace std;

wavespace::wavespace() 
{
	//fourier_space = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist/2+1)));
	//transform = fftw_plan_dft_r2c_2d(spat_dist, spat_dist, av, fourier_space, FFTW_ESTIMATE);
  	//transform_inverse = fftw_plan_dft_c2r_2d(spat_dist, spat_dist, fourier_space, av, FFTW_ESTIMATE);
	fourier_space = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	transform = fftw_plan_dft_2d(spat_dist, spat_dist, fourier_space, fourier_space, FFTW_FORWARD, FFTW_PATIENT);
  	transform_inverse = fftw_plan_dft_2d(spat_dist, spat_dist, fourier_space, fourier_space, FFTW_BACKWARD, FFTW_PATIENT);
	
	fourier_space_frec = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	frec_plan = fftw_plan_dft_2d(spat_dist, spat_dist, fourier_space_frec, fourier_space_frec, FFTW_FORWARD, FFTW_PATIENT);

	image = new QImage(spat_dist, spat_dist, QImage::Format_RGB32);
  //inicond();

  //SET PARAMETERS
	eps = 0.03;
	b = 0.001;
	a = 0.75;
	D = 1;
	dt = 0.08;
	periodicity = 0;
	L = 60;
	createEigenvalues(eig);
	spatialCorr = 13;
	Dv=0;
	dx=1;

  // ZERO PREVETIONS
	for (int xx=0; xx<(spat_dist); xx++)
    {
    	for (int yy=0; yy<(spat_dist); yy++)
    	{
      		inexfield[xx][yy]=0;
    	}
    }

	/* PLANS FOR THE FOURIERTRANFORM OF THE IFRK-METHOD */
	//FU2UV, FF12F1, iFUVn2iUVn, FF22F2, FF32F3, FF42F4

	UV = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	UVn= (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	iUVn=(fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	F1 = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	F2 = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	F3 = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));
	F4 = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) *(spat_dist*(spat_dist)));

	FUV2UV  = fftw_plan_dft_2d(spat_dist, spat_dist, UV, UV, FFTW_FORWARD, FFTW_PATIENT);
	iFUVn2iUVn = fftw_plan_dft_2d(spat_dist, spat_dist, UVn, iUVn, FFTW_BACKWARD, FFTW_PATIENT);
	FF12F1 = fftw_plan_dft_2d(spat_dist, spat_dist, F1, F1, FFTW_FORWARD, FFTW_PATIENT);
	FF22F2 = fftw_plan_dft_2d(spat_dist, spat_dist, F2, F2, FFTW_FORWARD, FFTW_PATIENT);
	FF32F3 = fftw_plan_dft_2d(spat_dist, spat_dist, F3, F3, FFTW_FORWARD, FFTW_PATIENT);
	FF42F4 = fftw_plan_dft_2d(spat_dist, spat_dist, F4, F4, FFTW_FORWARD, FFTW_PATIENT);
	iFUV2U = fftw_plan_dft_2d(spat_dist, spat_dist, UV, UV, FFTW_BACKWARD, FFTW_PATIENT);
}

wavespace::~wavespace() 
{
 // destructor
}

void wavespace::distribute_inxtblty(bool really, int noise_intencity) 
{
	if (really)
	{
	   srand(time(0));
	
		for (int xx=0; xx<(spat_dist); xx++)
			for (int yy=0; yy<(spat_dist); yy++)
			{
		  		int randbool = rand() % 1000;
				if (randbool<noise_intencity)
				{ 
					inexfield[xx][yy]=1;
					u_space[xx][yy]=0;
					v_space[xx][yy]=0;
				}
			}
	}
	else 
	{
		for (int xx=0; xx<(spat_dist); xx++)
			for (int yy=0; yy<(spat_dist); yy++)
					inexfield[xx][yy]=0;
	}
}

void wavespace::update_space(double intencity,int seed,QString solution_method,QString reaction_kinetics)
{
  double u_new[spat_dist][spat_dist];
  double v_new[spat_dist][spat_dist];

  //UPDATE INTERIOR POINTS
  #pragma omp parallel for schedule (static)
  for (int xx=0; xx<(spat_dist); xx++)
  {
    for (int yy=0; yy<(spat_dist); yy++)
    {
		u_new[xx][yy]=0;
        v_new[xx][yy]=0;

		if (xx==0 || yy==0 || xx==(spat_dist-1) || yy==(spat_dist-1)) continue;

		if (inexfield[xx][yy]) continue;
		
       	double gamma = (b+v_space[xx][yy])/a;

		passive_diffusion_5point(D,u_new[xx][yy],xx,yy);

		if (Dv>0)
		{
			passive_diffusion_5point_v(Dv,v_new[xx][yy],xx,yy);
		}
		
		if (solution_method=="explicit euler")
    	{
        	u_new[xx][yy]+= F_explicit(u_space[xx][yy],gamma);
		}
		else
		{
			u_new[xx][yy]+= F_implicit_standard(u_space[xx][yy],gamma) + u_new[xx][yy];	
		}
      	v_new[xx][yy]+=v_space[xx][yy]+dt*g_reaction(u_space[xx][yy],v_space[xx][yy],reaction_kinetics);
    }
  }
  
  	memcpy(u_space, u_new, sizeof (double) * spat_dist * spat_dist);
  	memcpy(v_space, v_new, sizeof (double) * spat_dist * spat_dist);

	//delete u_new;
	//delete[] v_new;

	/* add noise */	
	if ((int)M>0 && intencity>0)
	{
		add_noise(intencity,seed);
	}

}

void wavespace::spectral_update_space(double intencity, int seed,QString reaction_kinetics)
{
	gsl_rng * r = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(r,time(0)*seed);
	double r1;
	double sigm=(double)intencity/1;

	int k=0;
	
	/* Populate the input data in row-major order */

	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++, k++)
		{
    		fourier_space[k][0] = u_space[i][j];
			fourier_space[k][1] = 0;
		}
    }

	fftw_execute(transform);

	/* update diffusion in fourier space */
  #pragma omp parallel for schedule (static)
	for (int ww=0; ww<(spat_dist*(spat_dist)); ww++)
	{
		fourier_space[ww][0] = exp(-D*eig[ww]*dt)*fourier_space[ww][0];
		fourier_space[ww][1] = exp(-D*eig[ww]*dt)*fourier_space[ww][1];
		
		/*	applying noise
			x = ww%spat_dist, y = floor(i/spat_dist)	*/
		if ((ww%spat_dist)<M && (int)(floor(ww/spat_dist))<M)
		{
			r1 = gsl_ran_gaussian(r,1.0);

			double alpha=sqrt(exp(-eig[ww])*spatialCorr/pi);
			fourier_space[ww][0]+=sigm*dt*dt*r1*alpha;
			fourier_space[ww][1]+=sigm*dt*dt*r1*alpha;
		}
	}
	fftw_execute(transform_inverse);
	
	/* Re-populate the u_space-matrix and scale it*/
	k=0;	
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++, k++)
		{
    		u_space[i][j] = fourier_space[k][0]/(spat_dist*spat_dist);
		}
    }

	/* update reaction terms */
	#pragma omp parallel for schedule (static)
	for (int xx=0; xx<(spat_dist); xx++)
	{
		for (int yy=0; yy<(spat_dist); yy++)
		{
			if (inexfield[xx][yy]) continue;
		
			double gamma = (b+v_space[xx][yy])/a;
		
			u_space[xx][yy]=F_implicit_standard(u_space[xx][yy],gamma);
			//u_space[xx][yy]= F_explicit(u_space[xx][yy],gamma);  
			v_space[xx][yy]+=dt*g_reaction(u_space[xx][yy],v_space[xx][yy],reaction_kinetics);    	
		}
	}

	delete r;
}

void wavespace::eulerlawson_update_space(double intencity, int seed,QString reaction_kinetics)
{
	gsl_rng * r = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(r,time(0)*seed);
	double r1;

	int k=0;

	double sigm=(double)intencity/1;

	/* Populate the input data in row-major order */
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++, k++)
		{
    		fourier_space[k][0] = u_space[i][j];
			fourier_space[k][1] = 0;
		}
    }

	/* Transform u_space fourier_space=fft(u_space) */
	fftw_execute(transform);

	/* Determine the product u.*(I-u).*(u-gamma) */
	k=0;
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++, k++)
		{

    		fourier_space_frec[k][0] = u_space[i][j]*(1-u_space[i][j])*(u_space[i][j]-(v_space[i][j]+b)/a);
			fourier_space_frec[k][1] = 0;
		}
    }

	/* Transform the frec product to fourier space, */
	fftw_execute(frec_plan);

	/*oppdater euler_lawson step */
#pragma omp parallel for schedule (static)
	for (int ww=0; ww<(spat_dist*(spat_dist)); ww++)
	{
		double exphL=exp(-D*eig[ww]*dt);
		double phi=((exphL)-1)/exphL;
		phi = exphL;
		fourier_space[ww][0] = phi*fourier_space[ww][0]+dt/eps*phi*fourier_space_frec[ww][0];
		fourier_space[ww][1] = phi*fourier_space[ww][1]+dt/eps*phi*fourier_space_frec[ww][1];

		/*	applying noise
		x = ww%spat_dist, y = floor(i/spat_dist)	*/
		if ((ww%spat_dist)<M && (int)(floor(ww/spat_dist))<M)
		{
			r1 = gsl_ran_gaussian(r,1.0);

			double alpha=sqrt(exp(-eig[ww])*spatialCorr/pi);
			fourier_space[ww][0]+=sigm*sigm*sigm*dt/(L*L)*r1*alpha;
			fourier_space[ww][1]+=sigm*sigm*sigm*dt/(L*L)*r1*alpha;
		}
	}

	/* transform back to spatial space*/
	fftw_execute(transform_inverse);
	
	/* Re-populate the u_space-matrix and scale it*/
	k=0;

	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++,k++)
		{
    		u_space[i][j] = fourier_space[k][0]/(spat_dist*spat_dist);
			v_space[i][j] = v_space[i][j]+ dt*g_reaction(u_space[i][j],v_space[i][j],reaction_kinetics);
			//v_space[i][j] =v_space[i][j]+dt*(u_space[i][j]-v_space[i][j]);		
		}
    }

	delete r;

	clear_boundary();

}

void wavespace::inicond() 
{
  // SETTING INITIAL CONDITIONS
  for (int xx=0; xx<(spat_dist); xx++)
  {
    for (int yy=0; yy<(spat_dist); yy++)
    {
     if (xx>spat_dist/2)
      {
        v_space[xx][yy]=0.9;
      }
      else
      {
         v_space[xx][yy]=0;
      }

      if (yy>spat_dist/2)
      {
        u_space[xx][yy]=0.9;
      }
      else
      {
         u_space[xx][yy]=0;
      }
    }
  }
}

void wavespace::set_param(float new_eps, float new_b,float new_a, float new_D, float new_Dv, int new_L)
{
	eps=new_eps; b=new_b; a=new_a; D=new_D; Dv=new_Dv; L=new_L; dx=(float)L/spat_dist;
	createEigenvalues(eig);
};

double wavespace::g_reaction(double u,double v,QString reaction_kinetics)
{
	if (reaction_kinetics=="delayed")
	{
		if (v>=0) return u*u*u-v;
		else return abs(u*u*u-v);
	}	
	else if (reaction_kinetics=="standard")
    {
  		if (v>=0) return (u-v);
		else return abs(u-v);
	}
	else if (reaction_kinetics=="bareisw")
	{
		  double fu=0;

		  if (u<(0.3333))
		  { 
			fu=0-v;
		  }
		  else if ((0.3333)<=u && u<1)
		  {
			fu=1-6.75*u*(u-1)*(u-1)-v;
		  }
		  else
		  {
			fu=1-v;
		  }

		return fu;
	  }

	return 0;
}

double wavespace::F_implicit_standard(double u, double gamma)
{
   double F=0;

	if (u<=1)
	{
		if (u<=gamma)
		{
			F=u/(1-(dt/eps)*(1-u)*(u-gamma));
		}
		else
		{
			F=(u+(dt/eps)*u*(u-gamma))/(1+(dt/eps)*u*(u-gamma));
		}
	}
	else
	{
		F=1;
	}

    return F;
}

double wavespace::F_explicit(double u, double gamma)
{
   double F=0;

	if (u<=1)
	{
		F=u+(dt/eps)*u*(1-u)*(u-gamma);
	}
	else
	{
		F=u-abs((dt/eps)*u*(1-u)*(u-gamma));
	}
   
   return F;
}

void wavespace::passive_diffusion_5point(double Dd, double & u_new,int xx, int yy)
{

	if (xx>0 && xx<(spat_dist-1) && yy>0 && yy<(spat_dist-1))
	{
		u_new = dt*Dd*1/(dx*dx)*(u_space[xx+1][yy] +
			u_space[xx-1][yy] +
			u_space[xx][yy+1] +
			u_space[xx][yy-1] -
			4*u_space[xx][yy]);
		return;
	}

	if (periodicity)
	{
		if (xx==0)
		{
			u_new = dt*Dd*1/(dx*dx)*(u_space[xx+1][yy] +
				u_space[spat_dist-1][yy] +
				u_space[xx][yy+1] +
				u_space[xx][yy-1] -
				4*u_space[xx][yy]);
			
			return;
		}

		if (xx==(spat_dist-1))
		{
			u_new = dt*Dd*1/(dx*dx)*(u_space[0][yy] +
				u_space[xx-1][yy] +
				u_space[xx][yy+1] +
				u_space[xx][yy-1] -
				4*u_space[xx][yy]);
			return;
		}

		if (yy==0)
		{
			u_new = dt*Dd*1/(dx*dx)*(u_space[xx+1][yy] +
				u_space[xx-1][yy] +
				u_space[xx][yy+1] +
				u_space[xx][spat_dist-1] -
				4*u_space[xx][yy]);
			return;
		}

		if (yy==(spat_dist-1))
		{
			u_new = dt*Dd*1/(dx*dx)*(u_space[xx+1][yy] +
				u_space[xx-1][yy] +
				u_space[xx][0] +
				u_space[xx][yy-1] -
				4*u_space[xx][yy]);
			return;
		}
	}

}

void wavespace::passive_diffusion_5point_v(double Dd, double & v_new,int xx, int yy)
{

	if (xx>0 && xx<(spat_dist-1) && yy>0 && yy<(spat_dist-1))
	{
		v_new = dt*Dd*1/(dx*dx)*(v_space[xx+1][yy] +
			v_space[xx-1][yy] +
			v_space[xx][yy+1] +
			v_space[xx][yy-1] -
			4*v_space[xx][yy]);
		return;
	}

	if (periodicity)
	{
		if (xx==0)
		{
			v_new = dt*Dd*1/(dx*dx)*(v_space[xx+1][yy] +
				v_space[spat_dist-1][yy] +
				v_space[xx][yy+1] +
				v_space[xx][yy-1] -
				4*v_space[xx][yy]);
			
			return;
		}

		if (xx==(spat_dist-1))
		{
			v_new = dt*Dd*1/(dx*dx)*(v_space[0][yy] +
				v_space[xx-1][yy] +
				v_space[xx][yy+1] +
				v_space[xx][yy-1] -
				4*v_space[xx][yy]);
			return;
		}

		if (yy==0)
		{
			v_new = dt*Dd*1/(dx*dx)*(v_space[xx+1][yy] +
				v_space[xx-1][yy] +
				v_space[xx][yy+1] +
				v_space[xx][spat_dist-1] -
				4*v_space[xx][yy]);
			return;
		}

		if (yy==(spat_dist-1))
		{
			v_new = dt*Dd*1/(dx*dx)*(v_space[xx+1][yy] +
				v_space[xx-1][yy] +
				v_space[xx][0] +
				v_space[xx][yy-1] -
				4*v_space[xx][yy]);
			return;
		}
	}

}

void wavespace::passive_diffusion_9point(double & u_new,int xx, int yy)
{
	if (xx>1 && xx<(spat_dist-2) && yy>1 && yy<(spat_dist-2))
	{
		u_new = dt*D*1/(dx*dx)*(-5*u_space[xx][yy]
				-1/12*(u_space[xx-2][yy]+u_space[xx+2][yy]+u_space[xx][yy-2]+u_space[xx][yy+2])	
				+4/3*(u_space[xx-1][yy]+u_space[xx+1][yy]+u_space[xx][yy-1]+u_space[xx][yy+1]));	
		return;
	}
}

float wavespace::get_dt()
{
	return dt;
}

void wavespace::set_dt(float a)
{
	dt=a;
}

void wavespace::set_M(int a)
{
	M=a;
}

void wavespace::set_value(double u, double v, int x,int y)
{
	u_space[x][y]=u;
	printf("u_space[%i][%i]=%f \n",x,y,u);
	printf("v_space[%i][%i]=%f \n",x,y,v);
	v_space[x][y]=v;
}

void wavespace::createEigenvalues(double * lambda)
{
	int N = spat_dist;

	int k = N*N-1;

	double eig[N][N];

	for (int i = 0; i < N; i++)
	{
  		for (int j = 0; j < N; j++)
		{
			eig[i][j]=(pi/L)*(pi/L)*((i-N/2)*(i-N/2)+(j-N/2)*(j-N/2));
			//eig[i][j]=((i-N/2)*(i-N/2)+(j-N/2)*(j-N/2));			
		}
	}

	for (int i = 0; i < N; i++)
	{
  		for (int j = 0; j < N; j++)
		{		
			lambda[k]=eig[(i+N/2)%N][(j+N/2)%N];
			k--;	
		}
	}
}

void wavespace::clear()
{
	for (int xx=0; xx<(spat_dist); xx++)
  {
    for (int yy=0; yy<(spat_dist); yy++)
    {
      u_space[xx][yy]=0;
      v_space[xx][yy]=0;
    }
  }

}

void wavespace::clear_boundary()
{
	for (int xx=0; xx<(spat_dist); xx++)
	{
		u_space[xx][spat_dist-1]=0;
		u_space[xx][0]=0;
		u_space[spat_dist-1][xx]=0;
		u_space[0][xx]=0;
	
		v_space[xx][spat_dist-1]=0;
		v_space[xx][0]=0;
		v_space[spat_dist-1][xx]=0;
		v_space[0][xx]=0;	
	}
}

void wavespace::add_noise(double intencity, int seed)
{
	gsl_rng * r = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(r,time(0)*seed);
	double r1;
	double sigm=(double)intencity/1;

	int k=0;

	/* Populate the input data in row-major order */
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++, k++)
		{
    		fourier_space[k][0] = 0;
			fourier_space[k][1] = 0;
		}
    }

	/* update diffusion in fourier space */
  #pragma omp parallel for schedule (static)
	for (int ww=0; ww<(spat_dist*(spat_dist)); ww++)
	{		
		/*	applying noise
			x = ww%spat_dist, y = floor(i/spat_dist)	*/
		if ((ww%spat_dist)<M && (int)(floor(ww/spat_dist))<M)
		{
			r1 = sigm*sigm*dt/(L^2)*gsl_ran_gaussian(r,1.0);

			double alpha=sqrt(exp(-eig[ww])*spatialCorr/pi);
			fourier_space[ww][0]=sigm*r1*alpha;
			fourier_space[ww][1]=sigm*r1*alpha;
		}
	}

	fftw_execute(transform_inverse);

/* Re-populate the u_space-matrix and scale it*/
	k=0;

	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++)
		{
			k++;
    		u_space[i][j] += fourier_space[k][0]/(spat_dist*spat_dist);
		}
    }

	delete r;
}

#include <user_specified.h>

void wavespace::IFRK4(double intencity,int seed)
{
	double sigm=(double)intencity/1;
	int k=0;
	/* Populate the u_space in row-major order */
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				UV[k][0] = u_space[i][j];
				UV[k][1] = 0;
				k++;
			}
		}

	//-->fouriertransformer UV: UV=fft(u)
	fftw_execute(FUV2UV); 
	
	/* CALCULATE F1: F1=dt*fft2(1/eps*tildef(u,v,a,b,I)); */
		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F1[k][0] = dt/eps*(double)user_f(u_space[i][j],v_space[i][j]);
				F1[k][1] = 0;
				k++;
			}
		}
	
	//--> fouriertransformer F1: F1=fft(F1)
	fftw_execute(FF12F1); 

	/* UPDATE UVn=U2: U2 = exp(-D*eig*.5*dt).*(U+0.5*F1);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-D*eig[i]*0.5*dt)*(UV[i][0]+0.5*F1[i][0]);
			UVn[i][1] = 0;
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F2: F2=dt*exp(D*eig*.5*dt).*fft2(eps*tildef(ifft2(U2),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F2[k][0] = dt/eps*user_f(iUVn[k][0]/(spat_dist*spat_dist),v_space[i][j]);
				F2[k][1] = 0;
				k++;
			}
		}

		/* CALCULATE THE FOURIER TRANSFORM OF F2: F2 = FFT(F1) */ 
		fftw_execute(FF22F2); 
		
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
			F2[i][0] = exp(D*eig[i]*0.5*dt)*F1[i][0];
			F2[i][1] = exp(D*eig[i]*0.5*dt)*F1[i][1];
		}

	/* UPDATE UVn=U3: U3=exp(-D*eig*.5*dt).*(U+0.5*F2);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-D*eig[i]*0.5*dt)*(UV[i][0]+0.5*F2[i][0]);
			UVn[i][1] = exp(-D*eig[i]*0.5*dt)*(UV[i][1]+0.5*F2[i][1]);
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F3: F3=dt*exp(D*eig*.5*dt).*fft2(eps*tildef(ifft2(U3),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F3[k][0] = dt/eps*user_f(iUVn[k][0]/(spat_dist*spat_dist),v_space[i][j]);
				F3[k][1] = 0;
				k++;
			}
		}

		/* CALCULATE THE FOURIER TRANSFORM OF F3: F3 = FFT(F1) */
		fftw_execute(FF32F3); 		

	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
			F3[i][0] = exp(D*eig[i]*0.5*dt)*F3[i][0];
			F3[i][1] = exp(D*eig[i]*0.5*dt)*F3[i][1];
		}

	/* UPDATE UVn=U4: U4=exp(-D*eig*1*dt).*(U+1*F3);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-D*eig[i]*1*dt)*(UV[i][0]+1*F3[i][0]);
			UVn[i][1] = exp(-D*eig[i]*1*dt)*(UV[i][1]+1*F3[i][1]);
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F4: F4=dt*exp(D*eig*1*dt).*fft2(eps*tildef(ifft2(U4),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F4[k][0] = dt/eps*user_f(iUVn[k][0]/(spat_dist*spat_dist),v_space[i][j]);
				F4[k][1] = 0;
				k++;
			}
		}

		/* CALCULATE THE FOURIER TRANSFORM OF F4: F4 = FFT(F1) */
		fftw_execute(FF42F4); 
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
			F4[i][0] = exp(D*eig[i]*1*dt)*F4[i][0];
			F4[i][1] = exp(D*eig[i]*1*dt)*F4[i][1];
		}

	/* UPDATE U_{N+1}=U_N */
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
			UV[i][0] = exp(-D*eig[i]*1*dt)*(UV[i][0]+(F1[i][0]+2*F2[i][0]+2*F3[i][0]+F4[i][0])/6);
			UV[i][1] = exp(-D*eig[i]*1*dt)*(UV[i][1]+(F1[i][1]+2*F2[i][1]+2*F3[i][1]+F4[i][1])/6);
		}
	
	/* TRANSFORM UV BACK TO u */
	fftw_execute(iFUV2U);

	/* Re-populate the u_space-matrix and scale it*/
	k=0;
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++)
		{
    		u_space[i][j] = UV[k][0]/(spat_dist*spat_dist);
			k++;
		}
    }

	/* UPDATE V_N BY IFRK4 */

	/* Populate the u_space in row-major order */
		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				UV[k][0] = v_space[i][j];
				UV[k][1] = 0;
				k++;
			}
		}

	//-->fouriertransformer UV: UV=fft(u)
	fftw_execute(FUV2UV); 
	
	/* CALCULATE F1: F1=dt*fft2(1/eps*tildef(u,v,a,b,I)); */
		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F1[k][0] = dt*(double)user_g(u_space[i][j],v_space[i][j]);
				F1[k][1] = 0;
				k++;
			}
		}
	
	//--> fouriertransformer F1: F1=fft(F1)
	fftw_execute(FF12F1); 

	/* UPDATE UVn=U2: U2 = exp(-D*eig*.5*dt).*(U+0.5*F1);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-Dv*eig[i]*0.5*dt)*(UV[i][0]+0.5*F1[i][0]);
			UVn[i][1] = 0;
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F2: F2=dt*exp(D*eig*.5*dt).*fft2(eps*tildef(ifft2(U2),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F2[k][0] = dt*user_g(u_space[i][j],iUVn[k][0]/(spat_dist*spat_dist));
				F2[k][1] = 0;
				k++;
			}
		}
		

		/* CALCULATE THE FOURIER TRANSFORM OF F2: F2 = FFT(F1) */ 
		fftw_execute(FF22F2); 
		
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
				F2[i][0] = exp(Dv*eig[i]*0.5*dt)*F1[i][0];
				F2[i][1] = exp(Dv*eig[i]*0.5*dt)*F1[i][1];
		}

	/* UPDATE UVn=U3: U3=exp(-D*eig*.5*dt).*(U+0.5*F2);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-Dv*eig[i]*0.5*dt)*(UV[i][0]+0.5*F2[i][0]);
			UVn[i][1] = exp(-Dv*eig[i]*0.5*dt)*(UV[i][1]+0.5*F2[i][1]);
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F3: F3=dt*exp(D*eig*.5*dt).*fft2(eps*tildef(ifft2(U3),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F3[k][0] = dt*user_g(u_space[i][j],iUVn[k][0]/(spat_dist*spat_dist));
				F3[k][1] = 0;
				k++;
			}
		}

		/* CALCULATE THE FOURIER TRANSFORM OF F3: F3 = FFT(F1) */
		fftw_execute(FF32F3); 		

	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
				F3[i][0] = exp(Dv*eig[i]*0.5*dt)*F3[i][0];
				F3[i][1] = exp(Dv*eig[i]*0.5*dt)*F3[i][1];
		}

	/* UPDATE UVn=U4: U4=exp(-D*eig*1*dt).*(U+1*F3);*/
	#pragma omp parallel for schedule (static)
		for (int i=0; i<spat_dist*spat_dist; i++)
		{
			UVn[i][0] = exp(-Dv*eig[i]*1*dt)*(UV[i][0]+1*F3[i][0]);
			UVn[i][1] = exp(-Dv*eig[i]*1*dt)*(UV[i][1]+1*F3[i][1]);
		}

	// CALCULATE THE INVESRE OF UVn: iUVn=ifft(UVn)
	fftw_execute(iFUVn2iUVn);

	/* CALCULATE F4: F4=dt*exp(D*eig*1*dt).*fft2(eps*tildef(ifft2(U4),v,a,b,I)); */

		k=0;
		for (int i = 0; i < spat_dist; i++)
		{
	  		for (int j = 0; j < spat_dist; j++)
			{
				F4[k][0] = dt*user_g(u_space[i][j],iUVn[k][0]/(spat_dist*spat_dist));
				F4[k][1] = 0;
				k++;
			}
		}

		/* CALCULATE THE FOURIER TRANSFORM OF F4: F4 = FFT(F1) */
		fftw_execute(FF42F4); 
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
				F4[i][0] = exp(Dv*eig[i]*1*dt)*F4[i][0];
				F4[i][1] = exp(Dv*eig[i]*1*dt)*F4[i][1];
		}

	/* UPDATE U_{N+1}=U_N */
	#pragma omp parallel for schedule (static)
		for (int i = 0; i < spat_dist*spat_dist; i++)
		{
			UV[i][0] = exp(-Dv*eig[i]*1*dt)*(UV[i][0]+(F1[i][0]+2*F2[i][0]+2*F3[i][0]+F4[i][0])/6);
			UV[i][1] = exp(-Dv*eig[i]*1*dt)*(UV[i][1]+(F1[i][1]+2*F2[i][1]+2*F3[i][1]+F4[i][1])/6);
		}
	
	/* TRANSFORM UV BACK TO u */
	fftw_execute(iFUV2U);

	/* Re-populate the u_space-matrix and scale it*/
	k=0;
	for (int i = 0; i < spat_dist; i++)
	{
  		for (int j = 0; j < spat_dist; j++)
		{
    		v_space[i][j] = UV[k][0]/(spat_dist*spat_dist);
			k++;
		}
    }

	/* add noise */	
	if (M>0 && intencity>0)
	{
		add_noise(intencity,seed);
	}

	/* UPDATE v_n BY EXPLICIT EULER 
	for (int i = 0; i < spat_dist; i++)
	{
	  	for (int j = 0; j < spat_dist; j++)
		{
			v_space[i][j]=v_space[i][j]+dt*user_g(u_space[i][j],v_space[i][j]);
		}
	}*/ 

}
